/*-
 * Copyright (c) 2001 Nik Clayton
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
        "$Id: scr2png.c,v 1.3 2001/05/20 14:43:19 nik Exp $";
#endif /* not lint */

#include <sys/stat.h>
#include <sys/syslimits.h>
#include <sys/types.h>

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <gd.h>

#define	DEFAULT_FONT 	"/usr/share/syscons/fonts/cp850-8x16.fnt"
#define MAGIC		"SCRSHOT_"
#define VERSION		1

#define PAL_WIN 0
#define PAL_VGA 1

char palettes[2][48] = { 
/* Windows / xterm palette */
	0x00, 0x00, 0x00, /* Black */	0x00, 0x00, 0x80, /* Blue */
	0x00, 0x80, 0x00, /* Green */	0x00, 0x80, 0x80, /* Cyan */
	0x80, 0x00, 0x00, /* Red */	0x80, 0x00, 0x80, /* Magenta */
	0x80, 0x80, 0x00, /* Brown */	0xc0, 0xc0, 0xc0, /* White */

	/* Brights */
	0x80, 0x80, 0x80, /* Black */	0x00, 0x00, 0xff, /* Blue */
	0x00, 0xff, 0x00, /* Green */	0x00, 0xff, 0xff, /* Cyan */
	0xff, 0x00, 0x00, /* Red */	0xff, 0x00, 0xff, /* Magenta */
	0xff, 0xff, 0x00, /* Brown */	0xff, 0xff, 0xff, /* White */

/* Typical VGA Adapter palette */

	0x00, 0x00, 0x00, /* Black */	0x00, 0x00, 0xaa, /* Blue */
	0x00, 0xaa, 0x00, /* Green */	0x00, 0xaa, 0xaa, /* Cyan */
	0xaa, 0x00, 0x00, /* Red */	0xaa, 0x00, 0xaa, /* Magenta */
	0xaa, 0x55, 0x00, /* Brown */	0xaa, 0xaa, 0xaa, /* White */

	/* Brights */
	0x55, 0x55, 0x55, /* Black */	0x55, 0x55, 0xfe, /* Blue */
	0x55, 0xfe, 0x55, /* Green */	0x55, 0xfe, 0xfe, /* Cyan */
	0xfe, 0x55, 0x55, /* Red */	0xfe, 0x55, 0xfe, /* Magenta */
	0xfe, 0xfe, 0x55, /* Brown */	0xfe, 0xfe, 0xfe  /* White */
};

int
main(int argc, char *argv[])
{
	gdImagePtr	im;		/* Current image */
	gdFont		font;		/* Current font */
	int 		mx, my;		/* Maxmimum X and Y images positions */
	int 		cx, cy;		/* Current X and Y images positions */
	int 		colors[16]; 	/* Array of color RGB values */
	int 		curColor; 	/* Index into colors[] */
	int 		curChar; 	/* Current character to render */
	int 		ch;		/* Argument list option code */
	int		i;
	int		palette = PAL_WIN;
	char		header[BUFSIZ]; /* Buffer for header components */
	
        /* Full path to font filename */
	char 		fontfn[PATH_MAX] = "";

	/*
	 * Option processing
	 */
	while ((ch = getopt(argc, argv, "f:v")) != -1)
		switch (ch) {
		case 'f':
			strncpy(fontfn, optarg, PATH_MAX);
			break;

		case 'v':
			palette = PAL_VGA;
			break;
		}
  
	argc -= optind;
	argv += optind;

	if(fontfn[0] == '\0')
		strncpy(fontfn, DEFAULT_FONT, PATH_MAX);

	/*
	 * Load font
	 */
	syscons2gdfont(&font, fontfn);

	/*
	 * Read the SCRSHOT_ header, confirm the version, and read the image
	 * dimensions.
	 */
	bzero(header, BUFSIZ);
	
	if(fread(header, sizeof(unsigned char), 8, stdin) != 8)
		err(1, "fread() header bytes");

	if(strncmp(MAGIC, header, strlen(MAGIC)) != 0)
		err(1, "Expecting '%s', read '%s'", MAGIC, header);

	header[0] = getchar();
	if(header[0] != VERSION)
		fprintf(stderr, "Expecting version %d, read %d\n",
		    VERSION, header[0]);

	header[0] = getchar();
	if(header[0] == EOF)
		err(1, "Received EOF when reading header length");

	if(header[0] < 2)
		err(1, "Header length is less than 2 ('%d')", header[0]);
	
	mx = getchar();
	header[0]--;
	if(mx == EOF)
		err(1, "Received EOF when reading image width");

	my = getchar();
	header[0]--;
	if(my == EOF)
		err(1, "Received EOF when reading image height");
	
	/*
	 * Don't know what to do with any other data in the header,
	 * so skip it.
	 */
	while(header[0]-- != 0)
		getchar();
	
	/*
	 * Multiply the character cell dimensions by the font width and
	 * height to get the image size, in pixels.  Then create the image.
	 */
	mx *= font.w;
	my *= font.h;

	im = gdImageCreate(mx, my);

	if(im == NULL) 
	        err(1, "gdImageCreate() returned NULL");

	/*
	 * Initialise the color table
	 */
	for(i = 0; i < 16; i++) {
		colors[i] = gdImageColorAllocate(im, palettes[palette][3 * i],
						     palettes[palette][(3 * i) + 1],
						     palettes[palette][(3 * i) + 2]);
	}


	/*
	 * Read from STDIN, rendering on to the image
	 */
	cx = cy = 0;
	while(! feof(stdin)) {
		curChar = getchar();
		curColor = getchar();

		/* Fill in the background to this cell */
		gdImageFilledRectangle(im, cx, cy, cx + font.w, cy + font.h,
		    (curColor & 0xF0) >> 4);

		/* Add this character */
		gdImageChar(im, &font, cx, cy, curChar,
		    colors[curColor & 0x0F]);

		/* New line?  */
		cx += font.w;
		if(cx >= mx) {
			cx = 0;
			cy += font.h;
		}
	}

	/*
	 * Send the GIF to STDOUT and finish
	 */
	gdImagePng(im, stdout);
	gdImageDestroy(im);
	exit(0);
}

/*
 * Build a GD font from a syscons font. Handles binary or uuencoded fonts.
 *
 * Assumes;
 *
 *   The font file must exist.
 *
 *   The syscons font filename must be in the format foo-WxH where W is
 *   width of the font (almost certainly 8) and H is the height.
 *
 *   Each font file contains all 256 characters.
 *
 * Params;
 *
 *   gdFont *font       	Font to return
 *   char   *fn 		Syscons font to load from
 *
 * Chunks of this code based on src/usr.sbin/vidcontrol/vidcontrol.c
 */
int
syscons2gdfont(gdFont *font, char *fn)
{
	FILE 	*fp;
	int 	size, i;
	char 	*scfont;	/* Syscons font information */
	char 	*gdfont;	/* GD font information      */
  
	fp = fopen(fn, "r");
	if (fp == NULL)
		err(1, "%s", fn);
	
	font->nchars = 256;
	font->offset = 0;
	
	if (strstr(fn, "8x8") != NULL) {
		font->w = 8;
		font->h = 8;
	}
	else if (strstr(fn, "8x14") != NULL) {
		font->w = 8;
		font->h = 14;
	}
	else if (strstr(fn, "8x16") != NULL) {
		font->w = 8;
		font->h = 16;
	}
	else 
		err(1, "Cannot locate font size specification in filename %s",
		    fn);

	size = ((font->w * font->h) * font->nchars) / 8;

	/*
	 * Read (and decode if necessary) the syscons font file
	 */
	scfont = (char *)malloc(size);
	if (scfont == NULL)
		err(1, NULL);

	if (decode(fp, scfont) != size) {
		rewind(fp);

		if (fread(scfont, 1, size, fp) != size)
			err(1, "Bad font file: %s", fn);
	}	
  
	/*
	 * Syscons fonts are one bit per pixel, gd fonts are one byte per
	 * pixel. Expand as necessary.
	 */
	gdfont = (char *)malloc(size * 8);
	if(gdfont == NULL)
		err(1, NULL);
		   
	for(i = 0; i < size; i++) {
		gdfont[(i * 8)    ] = scfont[i] & 0x80;
		gdfont[(i * 8) + 1] = scfont[i] & 0x40;
		gdfont[(i * 8) + 2] = scfont[i] & 0x20;
		gdfont[(i * 8) + 3] = scfont[i] & 0x10;
		gdfont[(i * 8) + 4] = scfont[i] & 0x08;
		gdfont[(i * 8) + 5] = scfont[i] & 0x04;
		gdfont[(i * 8) + 6] = scfont[i] & 0x02;
		gdfont[(i * 8) + 7] = scfont[i] & 0x01;

	/*
	 * This inner loop would do the same thing, but I think the
	 * above is clearer.
	 *
	 *	for(j = 0; j < 8; j++) {
	 *	  gdfont[(i * 8) + j] = scfont[i] & (1 << (7 - j));
	 *	}
	 */
	}

	font->data = gdfont;
	return 0;
}
